from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from typing import Any

from avilla.core.trait.signature import ArtifactSignature


@dataclass(unsafe_hash=True)
class Scope:
    land: str
    mainline: str | None = None
    self: str | None = None


Namespace = dict[Scope | None, dict[ArtifactSignature, Any]]

ctx_namespace: ContextVar[Namespace] = ContextVar("ctx_namespace")

GLOBAL_SCOPE = Scope(land="avilla")

ctx_scope: ContextVar[Scope] = ContextVar("ctx_scope", default=GLOBAL_SCOPE)
ctx_prefix: ContextVar[str] = ContextVar("ctx_prefix", default="")


@contextmanager
def scope(land: str, mainline: str | None = None, self: str | None = None):
    token = ctx_scope.set(Scope(land, mainline, self))
    yield
    ctx_scope.reset(token)


def eval_dotpath(path: str, env: str = "") -> str:
    if path.startswith("."):
        if env:
            return ".".join(i for i in env.split(".") + path.split(".") if i)
        raise ValueError("cannot parse relative path without env pattern")
    return path


@contextmanager
def prefix(pattern: str):
    token = ctx_prefix.set(eval_dotpath(pattern, ctx_prefix.get()))
    yield
    ctx_prefix.reset(token)


@contextmanager
def wrap_namespace(initial: Namespace | None = None):
    namespace: Namespace = initial or {}
    token = ctx_namespace.set(namespace)
    yield namespace
    ctx_namespace.reset(token)


def raise_for_no_namespace(exception: type[Exception] = LookupError):
    if ctx_namespace.get(None) is None:
        raise exception("required available namespace to initize")


def get_current_namespace() -> dict[ArtifactSignature, Any]:
    return ctx_namespace.get().setdefault(ctx_scope.get(), {})



from __future__ import annotations

from asyncio import Future
from contextlib import suppress
from typing import Generic, TypeVar

T = TypeVar("T")

class Stream(Generic[T]):
    _receivers: list[Future[T]]
    closed: bool = False

    def __init__(self) -> None:
        self._receivers = []

    async def __iter__(self):
        if self.closed:
            raise RuntimeError("stream was closed.")
        while not self.closed:
            future = Future()
            self._receivers.append(future)
            try:
                yield await future
            finally:
                with suppress(ValueError):
                    self._receivers.remove(future)

    def push(self, value: T):
        for receiver in self._receivers:
            if not receiver.done() and not receiver.cancelled():
                receiver.set_result(value)

    def close(self):
        self.closed = True
        for receiver in self._receivers:
            if not receiver.done() and not receiver.cancelled():
                receiver.cancel()



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from graia.amnesia.message import MessageChain
from graia.broadcast.entities.dispatcher import BaseDispatcher

from avilla.core.event import AvillaEvent
from avilla.core.message import Message
from avilla.core.utilles.selector import Selector

if TYPE_CHECKING:
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface

    from avilla.core.account import AbstractAccount


class MessageReceived(AvillaEvent):
    message: Message

    @property
    def ctx(self) -> Selector:
        return self.message.sender

    def __init__(self, message: Message, account: AbstractAccount, time: datetime | None = None) -> None:
        self.message = message
        super().__init__(account, time=time)

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface: DispatcherInterface["MessageReceived"]):
            if interface.annotation is Message:
                return interface.event.message
            elif interface.annotation is MessageChain:
                return interface.event.message.content


class MessageEdited(AvillaEvent):
    message: Message
    operator: Selector
    past: MessageChain
    current: MessageChain

    @property
    def ctx(self):
        return self.operator

    def __init__(
        self,
        message: Message,
        operator: Selector,
        past: MessageChain,
        current: MessageChain,
        account: AbstractAccount,
        time: datetime | None = None,
    ) -> None:
        self.message = message
        self.operator = operator
        self.past = past
        self.current = current
        super().__init__(account, time=time)

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface: DispatcherInterface["MessageEdited"]):
            if interface.annotation is Message:
                return interface.event.message
            elif interface.annotation is MessageChain:
                return interface.event.current


class MessageRevoked(AvillaEvent):
    message: Selector
    operator: Selector

    @property
    def ctx(self):
        return self.operator

    def __init__(
        self,
        message: Selector,
        operator: Selector,
        account: AbstractAccount,
        time: datetime | None = None,
    ) -> None:
        self.message = message
        self.operator = operator
        super().__init__(account, time=time)

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface: DispatcherInterface["MessageRevoked"]):
            ...



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, ClassVar, cast

from graia.amnesia.message import MessageChain

from avilla.core.account import AbstractAccount

# from avilla.core.action.extension import ActionExtension
# from avilla.core.action.middleware import ActionMiddleware
from avilla.core.context import ctx_avilla, ctx_protocol
from avilla.core.event import AvillaEvent
from avilla.core.platform import Abstract, Land, Platform
from avilla.core.querier import ProtocolAbstractQueryHandler
from avilla.core.trait.context import Namespace
from avilla.core.utilles.event_parser import AbstractEventParser
from avilla.core.utilles.message_deserializer import MessageDeserializer
from avilla.core.utilles.message_serializer import MessageSerializer
from avilla.core.utilles.selector import Selector

if TYPE_CHECKING:
    from avilla.core.application import Avilla


class BaseProtocol(metaclass=ABCMeta):
    avilla: Avilla
    platform: ClassVar[Platform]
    message_serializer: MessageSerializer
    message_deserializer: MessageDeserializer

    # completion_rules: ClassVar[dict[str, dict[str, str]]] = cast(dict, MappingProxyType({}))
    # action_middlewares: list[ActionMiddleware] = []

    event_parser: ClassVar[AbstractEventParser]
    # action_executors: ClassVar[list[type[ProtocolActionExecutor]]] = cast(list, ())
    # resource_providers: ClassVar[dict[str, type[ProtocolResourceProvider]]] = cast(dict, MappingProxyType({}))
    # query_handlers: ClassVar[list[type[ProtocolAbstractQueryHandler]]] = cast(list, ())
    # extension_impls: ClassVar[dict[type[ActionExtension], ActionExtensionImpl]] = cast(dict, MappingProxyType({}))

    impl_namespace: ClassVar[Namespace]

    def __init__(self):
        ...

    @property
    def land(self):
        return self.platform[Land]

    @property
    def abstract(self):
        return self.platform[Abstract]

    """
    @property
    def resource_labels(self) -> set[str]:
        return {pattern.path_without_land for pattern in self.resource_providers.keys()}
    """

    @abstractmethod
    def ensure(self, avilla: Avilla) -> Any:
        ...

    def get_accounts(self, selector: Selector | None = None) -> list[AbstractAccount]:
        return self.avilla.get_accounts(selector=selector, land=self.platform[Land])

    def get_account(self, selector: Selector) -> AbstractAccount | None:
        return self.avilla.get_account(selector=selector, land=self.platform[Land])

    async def parse_message(self, data: Any) -> MessageChain:
        return MessageChain(await self.message_deserializer.parse_sentence(self, data))

    async def serialize_message(self, message: MessageChain) -> Any:
        return await self.message_serializer.serialize_chain(self, message)

    def post_event(self, event: AvillaEvent):
        with ctx_avilla.use(self.avilla), ctx_protocol.use(self):
            self.avilla.broadcast.postEvent(event)



from __future__ import annotations

import typing
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Concatenate,
    Generic,
    ParamSpec,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import Self

from avilla.core.utilles import identity
from avilla.core.utilles.selector import Selector

from ..cell import Cell, CellOf
from .signature import Impl, ImplDefaultTarget

if TYPE_CHECKING:
    from avilla.core.relationship import Relationship

from devtools import debug


class Trait:
    relationship: Relationship
    path: type[Cell] | CellOf | None
    target: Selector | None = None

    def __init__(
        self, relationship: Relationship, path: type[Cell] | CellOf | None = None, target: Selector | None = None
    ) -> None:
        self.relationship = relationship
        self.path = path
        self.target = target


_P = ParamSpec("_P")
_T = TypeVar("_T")

_P1 = ParamSpec("_P1")
_T1 = TypeVar("_T1")

_TboundTrait = TypeVar("_TboundTrait", bound=Trait)


class Fn(Generic[_P, _T]):
    __attr__: str
    schema: Callable[_P, Awaitable[_T]]

    @overload
    def __new__(cls, schema: Callable[Concatenate[_TboundTrait, _P1], Awaitable[_T1]]) -> Fn[_P1, _T1]:
        ...

    @overload
    def __new__(cls, schema: Callable[_P1, Awaitable[_T1]]) -> Fn[_P1, _T1]:
        ...

    def __new__(cls, schema: ...) -> ...:
        instance = super().__new__(cls)
        instance.schema = schema
        return instance

    def __set_name__(self, owner: type[Trait], attr: str):
        self.__attr__ = attr

    @overload
    def __get__(self, instance: Trait, owner: type[Trait]) -> FnWrapper[_P, _T]:
        ...

    @overload
    def __get__(self, instance: Any, owner: type[Trait] | None) -> Self:
        ...
    def __get__(self, instance: ..., owner: ...) -> Self | FnWrapper[_P, _T]:
        # sourcery skip: assign-if-exp, reintroduce-else, swap-if-expression
        if not isinstance(instance, Trait):
            return self
        return FnWrapper(instance, self)


class FnWrapper(Generic[_P, _T]):
    trait: Trait
    # call: TraitCall[_P, _T]

    def __init__(self, trait: Trait, fn: Fn[_P, _T]):
        self.trait = trait
        self.fn = fn

    async def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> _T:
        impl = self.trait.relationship._artifacts.get(
            Impl(
                self.fn,
                None,
                self.trait.path,
            )
        )
        if impl is None:
            raise NotImplementedError(
                f'"{identity(self.trait)}::{self.fn.__attr__}" '
                f'in "{identity(self.trait.relationship.protocol)}" '
                + (f'for target "{self.trait.target.path_without_land}"' if self.trait.target is not None else "")
                + "is not implemented"
            )
        impl = cast("Callable[Concatenate[Relationship, _P], Awaitable[_T]]", impl)
        return await impl(self.trait.relationship, *args, **kwargs)


class OrientedFn(Fn[_P, _T]):
    # 要求 Trait 里面要设置对象

    @overload
    def __new__(cls, schema: Callable[Concatenate[_TboundTrait, _P1], Awaitable[_T1]]) -> OrientedFn[_P1, _T1]:
        ...

    @overload
    def __new__(cls, schema: Callable[_P1, Awaitable[_T1]]) -> OrientedFn[_P1, _T1]:
        ...

    def __new__(cls, schema: ...) -> ...:
        return super().__new__(cls, schema)

    @overload
    def __get__(self, instance: Trait, owner: type[Trait]) -> OrientedFnWrapper[_P, _T]:
        ...

    @overload
    def __get__(self, instance: Any, owner: type[Trait] | None) -> Self:
        ...

    def __get__(self, instance: ..., owner: ...) -> ...:
        # sourcery skip: assign-if-exp, reintroduce-else
        if instance is None:
            return self
        return DirectFnWrapper(instance, self)


class OrientedFnWrapper(FnWrapper[_P, _T]):
    _args: tuple[Any, ...] | None = None
    _kwargs: dict[str, Any] | None = None

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs):
        self._args = args
        self._kwargs = kwargs
        return self

    def __await__(self):
        return self.__await_impl__().__await__()

    async def __await_impl__(self) -> _T:
        if self._args is None or self._kwargs is None:
            raise ValueError("cannot call without param in shape of schema")

        target = self.trait.target
        if target is None:
            raise ValueError(f'cannot call "{identity(self.trait)}::{self.fn.__attr__}" without target on cast')
        impl = self.trait.relationship._artifacts.get(
            Impl(
                self.fn,
                target.path_without_land,
                self.trait.path,
            )
        )
        if impl is None:
            debug(
                self.trait.relationship._artifacts,
                Impl(
                    self.fn,
                    target.path_without_land,
                    self.trait.path,
                ),
            )
            raise NotImplementedError(
                f'"{identity(self.trait)}::{self.fn.__attr__}" '
                f'in "{identity(self.trait.relationship.protocol)}" '
                f'for target "{target.path_without_land}" '
                "is not implemented"
            )
        impl = cast("Callable[Concatenate[Relationship, Selector, _P], Awaitable[_T]]", impl)
        return await impl(self.trait.relationship, target, *self._args, **self._kwargs)  # type: ignore


class DirectFn(Fn[_P, _T]):
    @overload
    def __new__(cls, schema: Callable[Concatenate[_TboundTrait, _P1], Awaitable[_T1]]) -> DirectFn[_P1, _T1]:
        ...

    @overload
    def __new__(cls, schema: Callable[_P1, Awaitable[_T1]]) -> DirectFn[_P1, _T1]:
        ...

    def __new__(cls, schema: ...) -> ...:
        return super().__new__(cls, schema)

    @overload
    def __get__(self, instance: Trait, owner: type[Trait]) -> DirectFnWrapper[_P, _T]:
        ...

    @overload
    def __get__(self, instance: Any, owner: type[Trait] | None) -> Self:
        ...

    def __get__(self, instance: ..., owner: ...) -> ...:
        # sourcery skip: assign-if-exp, reintroduce-else
        if instance is None:
            return self
        return DirectFnWrapper(instance, self)


class DirectFnWrapper(FnWrapper[_P, _T]):
    _args: tuple[Any, ...] | None = None
    _kwargs: dict[str, Any] | None = None
    target: Selector | None = None

    def to(self, target: Selector):
        self.target = target
        return self

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs):
        self._args = args
        self._kwargs = kwargs
        return self

    def __await__(self):
        return self.__await_impl__().__await__()

    async def __await_impl__(self) -> _T:
        if self._args is None or self._kwargs is None:
            raise ValueError("cannot call without param in shape of schema")

        target = self.target or self.trait.target
        if target is None:
            targetter = self.trait.relationship._artifacts.get(ImplDefaultTarget(self.trait.path, self.fn))
            if targetter is None:
                raise NotImplementedError(
                    f'"{self.trait.__class__.__name__}::{self.fn.__attr__}" '
                    "required a target, but no target given and no default getter implemented."
                )
            targetter = cast("Callable[[Relationship], Selector]", targetter)
            target = targetter(self.trait.relationship)
        impl = self.trait.relationship._artifacts.get(
            Impl(
                self.fn,
                target.path_without_land,
                self.trait.path,
            )
        )
        if impl is None:
            debug(
                self.trait.relationship._artifacts,
                Impl(
                    self.fn,
                    target.path_without_land,
                    self.trait.path,
                ),
            )
            raise NotImplementedError(
                f'"{self.trait.__class__.__name__}::{self.fn.__attr__}" '
                f'in "{self.trait.relationship.protocol.__class__.__name__}" '
                f'for target "{target.path_without_land}" '
                "is not implemented"
            )
        impl = cast("Callable[Concatenate[Relationship, Selector, _P], Awaitable[_T]]", impl)
        return await impl(self.trait.relationship, target, *self._args, **self._kwargs)  # type: ignore



from __future__ import annotations

from collections import ChainMap, deque
from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Iterable,
    TypeVar,
    cast,
    overload,
)

from graia.amnesia.message import Element, MessageChain, Text
from typing_extensions import Unpack

from avilla.core.account import AbstractAccount

# from avilla.core.action.middleware import ActionMiddleware
from avilla.core.cell import Cell, CellOf
from avilla.core.context import ctx_relationship
from avilla.core.message import Message
from avilla.core.request import Request
from avilla.core.resource import Resource
from avilla.core.skeleton.message import MessageTrait
from avilla.core.skeleton.request import RequestTrait
from avilla.core.skeleton.scene import SceneTrait
from avilla.core.trait import Trait
from avilla.core.trait.context import GLOBAL_SCOPE, Scope
from avilla.core.trait.recorder import Querier
from avilla.core.trait.signature import (
    ArtifactSignature,
    Check,
    CompleteRule,
    Pull,
    Query,
    ResourceFetch,
)
from avilla.core.utilles.selector import Selectable, Selector

if TYPE_CHECKING:
    from avilla.core.protocol import BaseProtocol

_T = TypeVar("_T")
_M = TypeVar("_M", bound=Cell)
_TboundTrait = TypeVar("_TboundTrait", bound=Trait)


async def _query_depth_generator(
    rs: Relationship,
    current: Querier,
    predicate: Selector,
    upper_generator: AsyncGenerator[Selector, None] | None = None,
):
    if upper_generator is not None:
        async for i in upper_generator:
            async for j in current(rs, i, predicate):
                yield j
    else:
        async for j in current(rs, None, predicate):
            yield j


@dataclass
class _MatchStep:
    upper: str
    start: int
    history: tuple[Query, ...]


def _find_querier_steps(artifacts: ChainMap[ArtifactSignature, Any], query_path: str) -> list[Query] | None:
    result: list[Query] | None = None
    frags: list[str] = query_path.split(".")
    queue: deque[_MatchStep] = deque([_MatchStep("", 0, ())])
    while queue:
        head: _MatchStep = queue.popleft()
        current_steps: list[str] = []
        for curr_frag in frags[head.start :]:
            current_steps.append(curr_frag)
            steps = ".".join(current_steps)
            full_path = f"{head.upper}.{steps}" if head.upper else steps
            head.start += 1
            if (query := Query(head.upper or None, steps)) in artifacts:
                if full_path == query_path:
                    if result is None or len(result) > len(head.history) + 1:
                        result = [*head.history, query]
                else:
                    queue.append(
                        _MatchStep(
                            full_path,
                            head.start,
                            head.history + (query,),
                        )
                    )
    return result


class Relationship:
    ctx: Selector
    mainline: Selector
    self: Selector
    via: Selector | None = None

    account: AbstractAccount
    cache: dict[str, Any]

    protocol: "BaseProtocol"

    _artifacts: ChainMap[ArtifactSignature, Any]
    # _middlewares: list[ActionMiddleware]

    def __init__(
        self,
        protocol: "BaseProtocol",
        ctx: Selector,
        mainline: Selector,
        selft: Selector,
        account: AbstractAccount,
        via: Selector | None = None,
        # middlewares: list[ActionMiddleware] | None = None,
    ) -> None:
        self.ctx = ctx
        self.mainline = mainline
        self.self = selft
        self.via = via
        self.account = account
        self.protocol = protocol
        # self._middlewares = middlewares or []
        self.cache = {"meta": {}}
        self._artifacts = ChainMap(
            self.protocol.impl_namespace.get(Scope(self.land.name, self.mainline.path_without_land, self.self.path_without_land), {}),
            self.protocol.impl_namespace.get(Scope(self.land.name, self.mainline.path_without_land), {}),
            self.protocol.impl_namespace.get(Scope(self.land.name, self=self.self.path_without_land), {}),
            self.protocol.impl_namespace.get(Scope(self.land.name), {}),
            self.protocol.impl_namespace.get(GLOBAL_SCOPE, {}),
            self.avilla.global_artifacts,
        )

    @property
    def avilla(self):
        return self.protocol.avilla

    @property
    def land(self):
        return self.protocol.land

    @property
    def is_resource(self) -> bool:
        # TODO: Auto inference for implementations of a "ctx"
        ...

    @property
    def app_current(self) -> Relationship | None:
        return ctx_relationship.get(None)

    async def query(self, pattern: Selector, with_land: bool = False):
        querier_steps: list[Query] | None = _find_querier_steps(
            self._artifacts, pattern.path if with_land else pattern.path_without_land
        )

        if querier_steps is None:
            raise NotImplementedError(f'cannot query "{pattern.path_without_land}" due to unknown step calc.')

        querier = cast("dict[Query, Querier]", {i: self._artifacts[i] for i in querier_steps})
        generators: list[AsyncGenerator[Selector, None]] = []

        past = []
        for k, v in querier.items():
            past.append(k.target)
            pred = pattern.mixin(".".join(past))
            current = _query_depth_generator(self, v, pred, generators[-1] if generators else None)
            generators.append(current)

        async for i in generators[-1]:
            if with_land:
                yield Selector.from_dict({"land": self.land.name, **i.pattern})
            else:
                yield i

    @overload
    async def check(self, *, check_via: bool = True) -> None:
        # 检查 Relationship 的存在性.
        # 如 Relationship 的存在性无法被验证为真, 则 Relationship 不成立, 抛出错误.
        ...

    @overload
    async def check(self, target: Selector, *, strict: bool = False, check_via: bool = True) -> bool:
        # 检查 target 相对于当前关系 Relationship 的存在性.
        # 注意, 这里是 "相对于当前关系", 如 Github 的项目若为 Private, 则对于外界/Amonymous来说是不存在的, 即使他从客观上是存在的.
        # 注意, target 不仅需要相对于当前关系是存在的, 由于关系本身处在一个 mainline 之中,
        # mainline 相当于工作目录或者是 docker 那样的应用容器, 后者是更严谨的比喻,
        # 因为有些操作**只能**在处于一个特定的 mainline 中才能完成, 这其中包含了访问并操作某些 target.
        # 在 strict 模式下, target 被视作包含 "仅在当前 mainline 中才能完成的操作" 的集合中,
        # 表示其访问或是操作必须以当前 mainline 甚至是 current(account) 为基础.
        # 如果存在可能的 via, 则会先检查 via 的存在性, 因为 via 是维系这段关系的基础.
        ...

    async def check(self, target: Selector | None = None, *, strict: bool = False, check_via: bool = True) -> bool | None:
        # FIXME: check this sentence again
        if check_via and self.via is not None:
            await self.check(self.via, strict=True, check_via=False)
        checker = self._artifacts.get(Check((target or self.ctx).path_without_land))
        if checker is None:
            raise NotImplementedError(
                f'cannot check existence & accessible of "{(target or self.ctx).path_without_land}" due to notimplemented checker'
            )
        result = checker(self, target)
        if strict and not result:
            raise ValueError(f"check failed on {target or self.ctx!r}")
        return result

    def complete(self, selector: Selector, with_land: bool = False):
        output_rule = self._artifacts.get(CompleteRule(selector.path_without_land))
        if output_rule is not None:
            output_rule = cast(str, output_rule)
            selector = Selector().mixin(output_rule, selector, self.ctx, self.mainline)
        if with_land and list(selector.pattern.keys())[0] != "land":
            selector.pattern = {"land": self.land.name, **selector.pattern}
        return selector

    async def fetch(self, resource: Resource[_T]) -> _T:
        fetcher = self._artifacts.get(ResourceFetch(type(resource)))
        if fetcher is None:
            raise NotImplementedError(
                f'cannot fetch "{resource.selector}" '
                + f' because no available fetch implement found in "{self.protocol.__class__.__name__}"'
            )
        return await fetcher(self, resource)

    async def pull(
        self,
        path: type[_M] | CellOf[Unpack[tuple[Any, ...]], _M],
        target: Selector | Selectable | None = None,
        *,
        flush: bool = False,
    ) -> _M:
        if isinstance(target, Selectable):
            target = target.to_selector()
        if target is not None:
            cached = self.cache["meta"].get(target)
            if cached is not None and path in cached:
                if flush:
                    del cached[path]
                elif not path.has_params():
                    return cached[path]

        puller = self._artifacts.get(Pull(target.path_without_land if target is not None else None, path))
        if puller is None:
            raise NotImplementedError(
                f'cannot pull "{path}"'
                + (f' for "{target.path_without_land}"' if target is not None else "")
                + f' because no available implement found in "{self.protocol.__class__.__name__}"'
            )
        puller = cast("Callable[[Relationship, Selector | None], Awaitable[_M]]", puller)
        result = await puller(self, target)
        if target is not None and not path.has_params():
            cached = self.cache["meta"].setdefault(target, {})
            cached[path] = result
        return result

    def cast(
        self,
        trait: type[_TboundTrait],
        path: type[Cell] | CellOf[Unpack[tuple[Any, ...]], Cell] | None = None,
        target: Selector | Selectable | None = None,
    ) -> _TboundTrait:
        if isinstance(target, Selectable):
            target = target.to_selector()
        return trait(self, path, target)

    def send_message(
        self, message: MessageChain | str | Iterable[str | Element], *, reply: Message | Selector | str | None = None
    ):
        if isinstance(message, str):
            message = MessageChain([Text(message)])
        elif not isinstance(message, MessageChain):
            message = MessageChain([]).extend(list(message))
        else:
            message = MessageChain([i if isinstance(i, Element) else Text(i) for i in message])

        if isinstance(reply, Message):
            reply = reply.to_selector()
        elif isinstance(reply, str):
            reply = self.mainline.copy().message(reply)

        return self.cast(MessageTrait).send(message, reply=reply)

    # TODO: more shortcuts, like `accept_request` etc.

    async def accept_request(self, request: Request | Selector):
        if isinstance(request, Request):
            request = request.to_selector()
        return await self.cast(RequestTrait, target=request).accept()

    async def reject_request(self, request: Request | Selector, reason: str | None = None, forever: bool = False):
        if isinstance(request, Request):
            request = request.to_selector()
        return await self.cast(RequestTrait, target=request).reject(reason, forever)

    async def cancel_request(self, request: Request | Selector):
        if isinstance(request, Request):
            request = request.to_selector()
        return await self.cast(RequestTrait, target=request).cancel()

    async def ignore_request(self, request: Request | Selector):
        if isinstance(request, Request):
            request = request.to_selector()
        return await self.cast(RequestTrait, target=request).ignore()

    async def leave_scene(self, scene: Selectable | Selector | None = None):
        if isinstance(scene, Selectable):
            scene = scene.to_selector()
        return await self.cast(SceneTrait, target=scene or self.mainline).leave()

    async def disband_scene(self, scene: Selectable | Selector | None = None):
        if isinstance(scene, Selectable):
            scene = scene.to_selector()
        return await self.cast(SceneTrait, target=scene or self.mainline).disband()

    async def remove_member(
        self, target: Selector, reason: str | None = None, scene: Selectable | Selector | None = None
    ):
        if isinstance(scene, Selectable):
            scene = scene.to_selector()
        return await self.cast(SceneTrait, target=scene or self.mainline).remove_member(target, reason)



from __future__ import annotations

from graia.amnesia.message import MessageChain

from avilla.core.trait import DirectFn, Fn, Trait
from avilla.core.utilles.selector import Selector


class MessageTrait(Trait):
    @DirectFn
    async def send(self, message: MessageChain, *, reply: Selector | None = None) -> Selector:
        ...

    @Fn
    async def revoke(self, message: Selector) -> None:
        ...

    @Fn
    async def edit(self, message: Selector, content: MessageChain) -> None:
        ...

    # MessageFetch => rs.pull(Message, target=...)



from __future__ import annotations

from asyncio import Future
from contextlib import suppress
from typing import Generic, TypeVar

T = TypeVar("T")

class Promise(Generic[T]):
    _futures: list[Future[T]]
    done: bool = False

    def __init__(self) -> None:
        self._futures = []

    def __await__(self):
        return self.__await_impl__().__await__()
    
    async def __await_impl__(self):
        future = Future()
        try:
            self._futures.append(future)
            return await future
        finally:
            with suppress(ValueError):
                self._futures.remove(future)

    def resolve(self, value: T):
        for future in self._futures:
            if not future.done() and not future.cancelled():
                future.set_result(value)
    
    def reject(self, exception: Exception):
        for future in self._futures:
            if not future.done() and not future.cancelled():
                future.set_exception(exception)
    
    def cancel(self):
        for future in self._futures:
            if not future.done() and not future.cancelled():
                future.cancel()



from __future__ import annotations

from collections import ChainMap
from itertools import filterfalse
from typing import TYPE_CHECKING, Any, Callable, Literal, Protocol, runtime_checkable

from typing_extensions import Self

from avilla.core.platform import Land

if TYPE_CHECKING:
    ...

MatchRule = Literal["any", "exact", "exist", "fragment", "startswith"]
Pattern = str | Callable[[str], bool]


class Selector:
    mode: MatchRule = "exact"
    pattern: dict[str, str]
    path_excludes: frozenset[str]
    referent: Any = None

    def __init__(self, *, mode: MatchRule = "exact", path_excludes: frozenset[str] = frozenset()) -> None:
        self.mode = mode
        self.path_excludes = path_excludes
        self.pattern = {}

    def __getattr__(self, name: str) -> Callable[[str], Self]:
        def wrapper(content: str) -> Self:
            self.pattern[name] = content
            return self

        return wrapper

    def __hash__(self) -> int:
        return hash("Selector") + hash(tuple(self.pattern.items()))

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Selector) and o.pattern == self.pattern

    def __contains__(self, key: str) -> bool:
        return key in self.pattern

    def __getitem__(self, key: str) -> str:
        return self.pattern[key]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(mode={self.mode}).{'.'.join(f'{k}({v})' for k, v in self.pattern.items())}"

    @property
    def empty(self) -> bool:
        return not self.pattern

    @property
    def path(self) -> str:
        return ".".join(filterfalse(self.path_excludes.__contains__, self.pattern))

    @property
    def path_without_land(self) -> str:
        return ".".join(filterfalse((self.path_excludes | {"land"}).__contains__, self.pattern))

    @classmethod
    def exist(cls) -> Self:
        return cls(mode="exist")

    @classmethod
    def any(cls) -> Self:
        return cls(mode="any")

    @classmethod
    def fragment(cls, *path_excludes: str) -> Self:
        return cls(mode="fragment", path_excludes=frozenset(path_excludes))

    @classmethod
    def from_dict(cls, pattern: dict) -> Self:
        instance = cls()
        instance.pattern = pattern
        return instance

    @property
    def latest_key(self) -> str:
        return list(self.pattern.keys())[-1]

    def land(self, land: Land | str):
        if isinstance(land, Land):
            land = land.name
        self.pattern["land"] = land
        return self

    def match(self, other: Selectable) -> bool:
        if not isinstance(other, Selector):
            if not isinstance(other, Selectable):
                return False
            other = other.to_selector()
        try:
            match = {
                "any": self._match_any,
                "exact": self._match_exact,
                "exist": self._match_exist,
                "fragment": self._match_fragment,
                "startswith": self._match_startswith,
            }[self.mode]
        except KeyError:
            raise ValueError(f"Unknown match rule: {self.mode}") from None

        return match(other)

    def _match_any(self, other: Selector) -> bool:
        return True

    def _match_exact(self, other: Selector) -> bool:
        return type(other) is Selector and self.path == other.path and self.pattern == other.pattern

    def _match_exist(self, other: Selector) -> bool:
        return set(self.pattern.items()).issubset(other.pattern.items())

    def _match_fragment(self, other: Selector) -> bool:
        fragment = list(self.pattern.items())
        full = list(other.pattern.items())

        try:
            start = full.index(fragment[0])
        except IndexError:
            return True
        except ValueError:
            return False

        return full[start : start + len(fragment)] == fragment

    def _match_startswith(self, other: Selector) -> bool:
        fragment = list(self.pattern.items())
        full = list(other.pattern.items())

        return all(fragment[i] == full[i] for i in range(len(fragment)))

    def mix(self, path: str, **pattern: str) -> Self:
        pattern = self.pattern | pattern
        instance = self.__class__(mode=self.mode, path_excludes=self.path_excludes)

        try:
            instance.pattern = {each: pattern[each] for each in path.split(".")}
        except KeyError:
            raise ValueError(f"given information cannot mix with {path}") from None

        return instance

    def mixin(self, path: str, *selectors: Self) -> Self:
        return self.mix(path, **ChainMap(*(x.pattern for x in selectors)))

    def appendix(self, key: str, value: str) -> Self:
        self.pattern[key] = value
        return self

    def copy(self) -> Self:
        instance = self.__class__(mode=self.mode, path_excludes=self.path_excludes)
        instance.pattern = self.pattern.copy()
        return instance

    def as_dyn(self) -> DynamicSelector:
        instance = DynamicSelector(mode=self.mode, path_excludes=self.path_excludes)
        for k, v in self.pattern.items():
            getattr(instance, k)(v)
        return instance

    def to_selector(self):
        return self

    def follows(self, pattern: str) -> bool:
        patterns: dict[str, str] = {}
        bracket_depth: int = 0
        path_buf: list[str] = []
        pattern_buf: list[str] = []
        for ch in pattern:
            if ch == "." and bracket_depth == 0:
                patterns["".join(path_buf)] = "".join(pattern_buf) or "*"
                path_buf.clear()
                pattern_buf.clear()
            if ch == "(":
                if bracket_depth:
                    pattern_buf.append(ch)
                bracket_depth += 1
            elif ch == ")":
                if not bracket_depth:
                    raise ValueError
                bracket_depth -= 1
                if bracket_depth:
                    pattern_buf.append(ch)
            else:
                (pattern_buf if bracket_depth else path_buf).append(ch)
        if path_buf:
            patterns["".join(path_buf)] = "".join(pattern_buf) or "*"
        return (self.path if "land" in patterns else self.path_without_land) == ".".join(patterns) and all(
            k in self.pattern and v in ["*", self.pattern[k]] for k, v in patterns.items()
        )

    def set_referent(self, referent: Any) -> Self:
        self.referent = referent
        return self


class DynamicSelector(Selector):
    pattern: dict[str, Pattern]

    def __getattr__(self, name: str, /):
        def wrapper(content: Pattern | Literal["*"]):
            if content == "*":
                content = lambda _: True

            self.pattern[name] = content
            return self

        return wrapper

    def __hash__(self) -> int:
        raise TypeError("Dynamic Selector is unhashable.")

    __getitem__: Callable[[str], Pattern]

    @classmethod
    def way(cls, path: str) -> Selector:
        instance = cls()
        instance.pattern = {i: lambda _: True for i in path.split(".")}
        return instance

    def _match_exact(self, other: Selector) -> bool:
        if isinstance(other, DynamicSelector):
            raise TypeError("Can't match dynamic selector with another dynamic selector")
        for k in other.pattern:
            if k not in self.pattern:
                return False
            own_pattern = self.pattern[k]
            if (
                callable(own_pattern)
                and not own_pattern(other.pattern[k])
                or not callable(own_pattern)
                and own_pattern != other.pattern[k]
            ):
                return False
        return True

    def _match_exist(self, other: Selector) -> bool:
        for a, b in ((self.pattern[path], other.pattern[path]) for path in self.pattern):
            if callable(a):
                if callable(b):
                    raise TypeError("Can't partially match dynamic selector with another dynamic selector")
                elif not a(b):
                    return False
            elif a != b:
                return False

        return True

    def _match_fragment(self, other: Selector) -> bool:
        fragment = list(self.pattern)
        full = list(other.pattern)

        try:
            start = full.index(fragment[0])
        except IndexError:
            return True
        except ValueError:
            return False

        full = full[start : start + len(fragment)]
        if fragment != full:
            return False

        for a, b in ((self.pattern[path], other.pattern[path]) for path in fragment):
            if callable(a):
                if callable(b):
                    raise TypeError("Can't partially match dynamic selector with another dynamic selector")
                elif not a(b):
                    return False
            elif a != b:
                return False

        return True

    def _match_startswith(self, other: Selector) -> bool:
        if not other.path.startswith(self.path):
            return False

        for a, b in ((self.pattern[path], other.pattern[path]) for path in self.pattern):
            if callable(a):
                if callable(b):
                    raise TypeError("Can't partially match dynamic selector with another dynamic selector")
                elif not a(b):
                    return False
            elif a != b:
                return False

        return True


@runtime_checkable
class Selectable(Protocol):
    def to_selector(self) -> Selector:
        ...



from __future__ import annotations

from avilla.core.trait import OrientedFn, Trait


class RequestTrait(Trait):
    @OrientedFn
    async def accept(self):
        ...

    @OrientedFn
    async def reject(self, reason: str | None = None, forever: bool = False):
        ...

    @OrientedFn
    async def cancel(self):
        ...

    @OrientedFn
    async def ignore(self):
        ...



import inspect
from typing import Any, Awaitable, Callable, Coroutine, TypeVar, overload

T = TypeVar("T")


@overload
async def run_always_await(callable: Callable[..., T], *args, **kwargs) -> T:
    ...


@overload
async def run_always_await(callable: Callable[..., Awaitable[T]], *args, **kwargs) -> T:
    ...


@overload
async def run_always_await(callable: Callable[..., Coroutine[None, None, T]], *args, **kwargs) -> T:
    ...


async def run_always_await(callable, *args, **kwargs):
    obj = callable(*args, **kwargs)
    while inspect.isawaitable(obj):
        obj = await obj
    return obj


def identity(obj: Any) -> str:
    return obj.__name__ if isinstance(obj, type) else obj.__class__.__name__



from __future__ import annotations

from typing import TYPE_CHECKING

from launart import Launart, Service

from avilla.core.event.lifecycle import (
    ApplicationClosed,
    ApplicationClosing,
    ApplicationPreparing,
    ApplicationReady,
)

if TYPE_CHECKING:
    from .application import Avilla


class AvillaService(Service):
    id = "avilla.service"
    supported_interface_types = set()

    avilla: Avilla

    def __init__(self, avilla: Avilla):
        self.avilla = avilla
        super().__init__()

    @property
    def required(self) -> set[str]:
        return set()

    @property
    def stages(self):
        return {"preparing", "blocking", "cleanup"}

    def get_interface(self, interface_type):
        ...

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            await self.avilla.broadcast.postEvent(ApplicationPreparing(self.avilla))

        self.avilla.broadcast.postEvent(ApplicationReady(self.avilla))

        async with self.stage("blocking"):
            ...  # TODO: 先放着.

        async with self.stage("cleanup"):
            self.avilla.broadcast.postEvent(ApplicationClosing(self.avilla))

        self.avilla.broadcast.postEvent(ApplicationClosed(self.avilla))



from __future__ import annotations

import inspect
from typing import (
    TYPE_CHECKING,
    AsyncGenerator,
    AsyncIterator,
    Callable,
    ClassVar,
    Generic,
    TypeVar,
)

from typing_extensions import Self

from avilla.core.utilles.selector import Selector

if TYPE_CHECKING:
    from avilla.core.protocol import BaseProtocol
    from avilla.core.relationship import Relationship


def query(target: str):
    def wrapper(func):
        func.__query_key__ = target
        return func

    return wrapper


class AbstractQueryHandler:
    prefix: str | None
    queriers: ClassVar[
        dict[
            str,
            Callable[[Self, Relationship, Selector, Callable[[Selector], bool]], AsyncGenerator[Selector, None]],
        ]
    ] = {}

    def __init_subclass__(cls, prefix: str | None = None):
        super().__init_subclass__()
        cls.queriers = {}
        cls.prefix = prefix
        for mro in reversed(inspect.getmro(cls)):
            if issubclass(mro, AbstractQueryHandler):
                cls.queriers.update(mro.queriers)
        members = inspect.getmembers(cls, predicate=inspect.isfunction)
        for _, value in members:
            if hasattr(value, "__query_key__"):
                cls.queriers[value.__query_key__] = value


TProtocol = TypeVar("TProtocol", bound="BaseProtocol")


class ProtocolAbstractQueryHandler(AbstractQueryHandler, Generic[TProtocol]):
    protocol: TProtocol

    def __init__(self, protocol: TProtocol) -> None:
        self.protocol = protocol

    def __init_subclass__(cls, prefix: str | None = None):
        super().__init_subclass__(prefix)



from __future__ import annotations

from avilla.core.trait import Fn, OrientedFn, Trait
from avilla.core.utilles.selector import Selector


class NickTrait(Trait):
    @OrientedFn
    async def set_name(self, name: str) -> None:
        ...

    @OrientedFn
    async def set_nickname(self, nickname: str) -> None:
        ...

    @OrientedFn
    async def unset_nickname(self) -> None:
        ...

    # TODO: recheck badge operators

    @Fn
    async def set_badge(self, target: Selector, badge: str) -> None:
        ...

    @Fn
    async def unset_badge(self, target: Selector) -> None:
        ...



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Coroutine, Generic, TypeVar

from graia.amnesia.message.element import Element
from graia.broadcast.utilles import run_always_await
from loguru import logger
from typing_extensions import Self

from avilla.core.account import AbstractAccount
from avilla.core.event import AvillaEvent

if TYPE_CHECKING:
    from avilla.core.account import AbstractAccount
    from avilla.core.protocol import BaseProtocol


def event(event_type: str):
    def wrapper(func):
        func.__event_parser__ = event_type
        return func

    return wrapper


_P = TypeVar("_P", bound="BaseProtocol")


class AbstractEventParser(ABC, Generic[_P]):
    event_parser: dict[str, Callable[[Self, _P, AbstractAccount, dict], Coroutine[None, None, AvillaEvent | None]]] = {}

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls.event_parser = {}
        for mro in reversed(inspect.getmro(cls)):
            if issubclass(mro, AbstractEventParser):
                cls.event_parser.update(mro.event_parser)
        members = inspect.getmembers(cls)
        for _, value in members:
            if callable(value) and getattr(value, "__event_parser__", False):
                cls.event_parser[value.__event_parser__] = value

    @abstractmethod
    def get_event_type(self, raw: dict) -> str:
        ...

    async def parse_event(
        self, protocol: _P, account: AbstractAccount, raw: dict, *, error: bool = False
    ) -> AvillaEvent | None:
        event_type = self.get_event_type(raw)
        deserializer = self.event_parser.get(event_type)
        if deserializer is None:
            if error:
                raise NotImplementedError(f"Event type {event_type} is not supported.")
            logger.warning(f"Event type {event_type} is not supported by {self.__class__.__name__}", raw)
            return
        return await deserializer(self, protocol, account, raw)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Generic, TypeVar

from graia.amnesia.message import MessageChain
from graia.amnesia.message.element import Element
from graia.broadcast.utilles import run_always_await
from typing_extensions import Self

if TYPE_CHECKING:
    from avilla.core.protocol import BaseProtocol


def element(element_type: type[Element]):
    def wrapper(func: Callable):
        func.__element_serializer__ = element_type
        return func

    return wrapper


_P = TypeVar("_P", bound="BaseProtocol")


class MessageSerializer(Generic[_P]):
    element_serializer: dict[type[Element], Callable[[Self, _P, Any], dict | Coroutine[None, None, dict]]] = {}

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls.element_serializer = {}
        for mro in reversed(inspect.getmro(cls)):
            if issubclass(mro, MessageSerializer):
                cls.element_serializer.update(mro.element_serializer)
        members = inspect.getmembers(cls)
        for _, value in members:
            if callable(value) and getattr(value, "__element_serializer__", False):
                cls.element_serializer[value.__element_serializer__] = value

    async def serialize_element(self, protocol: _P, element: Element) -> dict:
        if type(element) not in self.element_serializer:
            raise NotImplementedError(f"Element type {type(element)} is not supported.")
        return await run_always_await(self.element_serializer[type(element)], self, protocol, element)  # type: ignore

    async def serialize_chain(self, protocol: _P, message: MessageChain):
        return [await self.serialize_element(protocol, element) for element in message.content]



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Generic, Iterable, TypeVar

from graia.amnesia.message.element import Element
from typing_extensions import Self

from avilla.core.elements import Unknown
from avilla.core.utilles import run_always_await

if TYPE_CHECKING:
    from avilla.core.protocol import BaseProtocol


def deserializer(element_type: str):
    def wrapper(func):
        func.__element_deserializer__ = element_type
        return func

    return wrapper


_P = TypeVar("_P", bound="BaseProtocol")


class MessageDeserializer(ABC, Generic[_P]):
    element_deserializer: dict[
        str, Callable[[Self, _P, dict], Coroutine[None, None, Element]] | Callable[[Self, _P, dict], Element]
    ] = {}
    ignored_types: set[str] = set()

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        cls.element_deserializer = {}
        for mro in reversed(inspect.getmro(cls)):
            if issubclass(mro, MessageDeserializer):
                cls.element_deserializer.update(mro.element_deserializer)
        members = inspect.getmembers(cls)
        for _, value in members:
            if callable(value) and getattr(value, "__element_deserializer__", False):
                cls.element_deserializer[value.__element_deserializer__] = value

    def split_message(self, data: list) -> Iterable[dict]:
        yield from data

    @abstractmethod
    def get_element_type(self, raw: dict) -> str:
        ...

    async def parse_element(self, protocol: _P, raw: dict) -> Element:
        element_type = self.get_element_type(raw)
        deserializer = self.element_deserializer.get(element_type)
        if deserializer is None:
            return Unknown(type=element_type, raw_data=raw)
        return await run_always_await(deserializer, self, protocol, raw)  # type: ignore

    async def parse_sentence(self, protocol: _P, data: list) -> list[Any]:
        return [
            await self.parse_element(protocol, raw)
            for raw in self.split_message(data)
            if self.get_element_type(raw) not in self.ignored_types
        ]



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from avilla.core.platform import Land
from avilla.core.utilles.selector import Selector

if TYPE_CHECKING:
    from avilla.core.protocol import BaseProtocol
    from avilla.core.relationship import Relationship


@dataclass
class AbstractAccount(ABC):
    id: str
    land: Land
    protocol: BaseProtocol

    def __init__(self, id: str, protocol: BaseProtocol, land: Land | None = None):
        self.id = id
        self.land = land or protocol.land
        self.protocol = protocol

    @abstractmethod
    async def get_relationship(self, target: Selector, *, via: Selector | None = None) -> Relationship:
        ...

    @abstractmethod
    async def call(self, endpoint: str, params: dict[str, Any] | None = None) -> Any:
        ...

    def get_self_relationship(self):
        return Relationship(
            self.protocol, self.to_selector(), Selector().land(self.land.name), self.to_selector(), self
        )

    @property
    def available(self) -> bool:
        return True

    def to_selector(self) -> Selector:
        return Selector().land(self.land.name).account(self.id)

    def is_anonymous(self) -> bool:
        return self.id == "anonymous"

    def __eq__(self, other: AbstractAccount):
        return (
            self.__class__ is other.__class__
            and self.id == other.id
            and self.land == other.land
            and self.protocol == other.protocol
        )



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING

from avilla.core.cell import Cell
from avilla.core.platform import Land

if TYPE_CHECKING:
    from avilla.core.account import AbstractAccount
    from avilla.core.utilles.selector import Selector


@dataclass
class Request(Cell):
    id: str
    land: Land
    mainline: Selector
    sender: Selector
    account: AbstractAccount
    time: datetime

    request_type: str | None = None
    operator: Selector | None = None
    response: str | None = None

    @property
    def solved(self) -> bool:
        return self.response is not None

    @property
    def pending(self) -> bool:
        return self.response is None

    @property
    def accepted(self) -> bool:
        return self.response == "accept"

    @property
    def rejected(self) -> bool:
        return self.response == "reject"

    @property
    def cannelled(self) -> bool:
        return self.response == "cancel"

    @property
    def ignored(self) -> bool:
        return self.response == "ignore"

    def to_selector(self) -> Selector:
        request_id = self.id
        if self.request_type is not None:
            request_id = f"{self.request_type}:{request_id}"
        return self.mainline.copy().request(request_id)



from __future__ import annotations

from contextvars import ContextVar
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, TypeVar, overload
from weakref import WeakKeyDictionary

from typing_extensions import Self, TypeVarTuple, Unpack

if TYPE_CHECKING:
    from avilla.core.relationship import Relationship
    from avilla.core.utilles.selector import Selector

T = TypeVar("T")  # 返回值

_Meta_A = TypeVar("_Meta_A", bound="Cell")

_Meta_B = TypeVar("_Meta_B", bound="Cell")

TVT = TypeVarTuple("TVT")


class CellMeta(type):
    @overload
    def __rshift__(cls: type[_Meta_A], other: type[_Meta_B]) -> CellOf[_Meta_A, _Meta_B]:
        # sourcery skip: instance-method-first-arg-name
        ...

    @overload
    def __rshift__(cls: type[_Meta_A], other: CellOf[Unpack[TVT]]) -> CellOf[_Meta_A, Unpack[TVT]]:
        # sourcery skip: instance-method-first-arg-name
        ...

    def __rshift__(cls: Any, other: type[Cell] | CellOf) -> CellOf:
        # sourcery skip: instance-method-first-arg-name
        cells = (cls, other) if isinstance(other, type) else (cls, *other.cells)
        return CellOf(cells)

    """
    def __repr__(cls) -> str:
        # sourcery skip: instance-method-first-arg-name
        return cls.__name__
    """


METACELL_PARAMS_CTX: WeakKeyDictionary[type[Cell], ContextVar[dict[str, Any] | None]] = WeakKeyDictionary()


@dataclass
class Cell(Generic[T], metaclass=CellMeta):
    describe: type[Self] | CellOf  # CellCompose 的行为和单次的行为一样.

    @classmethod
    @property
    def _param_ctx(cls):
        return METACELL_PARAMS_CTX[cls]

    def __init_subclass__(cls) -> None:
        METACELL_PARAMS_CTX[cls] = ContextVar(f"$MetadataParam${cls.__module__}::{cls.__qualname__}", default=None)

    @classmethod
    def get_default_target(cls, relationship: Relationship) -> Selector | None:
        return relationship.ctx.to_selector()

    @classmethod
    def get_params(cls) -> dict[str, Any] | None:
        return cls._param_ctx.get()

    @classmethod
    def set_params(cls, params: dict[str, Any]):
        target = cls._param_ctx.get()
        if target is None:
            target = {}
            cls._param_ctx.set(target)
        target.update(params)

    @classmethod
    def clear_params(cls) -> None:
        cls._param_ctx.set(None)

    @classmethod
    def has_params(cls) -> bool:
        return cls._param_ctx.get() is not None


Ts = TypeVarTuple("Ts")

_M = TypeVar("_M", bound=Cell)

_M_k = TypeVar("_M_k", bound=Cell)


class CellOf(Generic[Unpack[TVT]]):
    cells: tuple[type[Cell], ...]

    def __init__(self, cells: tuple[type[Cell], ...]) -> None:
        self.cells = cells

    def get_default_target(self, relationship: Relationship) -> Selector | None:
        return self.cells[0].get_default_target(relationship)

    @overload
    def __rshift__(self: CellOf[Unpack[TVT]], other: type[_M]) -> CellOf[Unpack[TVT], _M]:
        ...

    @overload
    def __rshift__(self: CellOf[Unpack[TVT]], other: CellOf[Unpack[Ts]]) -> CellOf[Unpack[TVT], Unpack[Ts]]:
        ...

    def __rshift__(self, other: type[Cell] | CellOf) -> CellOf:
        if not isinstance(other, (type, CellOf)):
            raise TypeError(f"{other.__class__} is not allowed.")
        cells = (*self.cells, other) if isinstance(other, type) else (*self.cells, *other.cells)
        return CellOf(cells)

    def __repr__(self) -> str:
        cells_repr = " >> ".join(cell.__name__ for cell in self.cells)
        return f"CellOf[{cells_repr}]"

    def __hash__(self) -> int:
        return hash(self.cells) + hash("CellOf")

    def __eq__(self, o: object) -> bool:
        return isinstance(o, CellOf) and o.cells == self.cells

    def clear_params(self) -> None:
        for cell in self.cells:
            cell.clear_params()

    def has_params(self) -> bool:
        return any(cell.has_params() for cell in self.cells)


_DeriveBack = CellOf[Unpack[tuple[Any, ...]], _M_k]



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Any

from graia.broadcast.entities.event import Dispatchable

if TYPE_CHECKING:
    from avilla.core.account import AbstractAccount
    from avilla.core.utilles.selector import Selector

    from ..cell import Cell, CellOf


class AvillaEvent(Dispatchable, metaclass=ABCMeta):
    account: AbstractAccount
    time: datetime

    extra: dict[Any, Any]

    def __init__(
        self, account: AbstractAccount, *, extra: dict[Any, Any] | None = None, time: datetime | None = None
    ) -> None:
        self.account = account
        self.extra = extra or {}
        self.time = time or datetime.now()

    @property
    @abstractmethod
    def ctx(self) -> Selector:
        ...

    def get_via(self) -> Selector | None:
        ...


@dataclass
class MetadataModify:
    describe: type[Cell] | CellOf
    field: str
    current: Any
    past: Any | None = None


class MetadataModified(AvillaEvent):
    ctx: Selector
    modifies: list[MetadataModify]
    operator: Selector | None = None

    def __init__(
        self,
        ctx: Selector,
        modifies: list[MetadataModify],
        account: AbstractAccount,
        operator: Selector | None = None,
        time: datetime | None = None,
        extra: dict[Any, Any] | None = None,
    ):
        self.ctx = ctx
        self.modifies = modifies
        self.operator = operator
        super().__init__(account, time=time, extra=extra)


class RelationshipCreated(AvillaEvent):
    ctx: Selector
    via: Selector | None
    # 用 via 同时表示两个方向的关系.(自发行为和被动行为)
    # 自发行为就是 None, 被动行为反之

    def __init__(
        self,
        ctx: Selector,
        account: AbstractAccount,
        time: datetime | None = None,
        via: Selector | None = None,
    ):
        self.ctx = ctx
        self.via = via
        super().__init__(account, time=time)

    def get_via(self) -> Selector | None:
        return self.via


class RelationshipDestroyed(AvillaEvent):
    ctx: Selector
    via: Selector | None

    def __init__(
        self,
        ctx: Selector,
        account: AbstractAccount,
        time: datetime | None = None,
        via: Selector | None = None,
    ):
        self.ctx = ctx
        self.via = via
        super().__init__(account, time=time)

    def get_via(self) -> Selector | None:
        return self.via



from __future__ import annotations

from inspect import isclass
from types import TracebackType
from typing import TYPE_CHECKING

from graia.broadcast.entities.dispatcher import BaseDispatcher

from avilla.core.account import AbstractAccount
from avilla.core.cell import Cell
from avilla.core.context import ctx_protocol, ctx_relationship
from avilla.core.event import AvillaEvent
from avilla.core.relationship import Relationship

if TYPE_CHECKING:
    from graia.broadcast.interfaces.dispatcher import DispatcherInterface

    from avilla.core.application import Avilla


class AvillaBuiltinDispatcher(BaseDispatcher):
    avilla: Avilla

    def __init__(self, avilla: Avilla) -> None:
        self.avilla = avilla

    async def catch(self, interface: DispatcherInterface[AvillaEvent]):
        from avilla.core.application import Avilla

        if interface.annotation is Avilla:
            return self.avilla
        elif interface.annotation in self.avilla._protocol_map:
            return self.avilla._protocol_map[interface.annotation]
        elif isinstance(interface.event, AvillaEvent):
            if isclass(interface.annotation) and issubclass(interface.annotation, AbstractAccount):
                rs: Relationship = interface.local_storage["relationship"]
                return interface.event.account


class RelationshipDispatcher(BaseDispatcher):
    @staticmethod
    async def beforeExecution(interface: DispatcherInterface[AvillaEvent]):
        protocol = ctx_protocol.get()
        if protocol is not None:
            rs = await interface.event.account.get_relationship(interface.event.ctx, via=interface.event.get_via())
            token = ctx_relationship.set(rs)
            rs.cache["meta"] = interface.event.extra.get("meta", {})
            interface.local_storage["relationship"] = rs
            interface.local_storage["_ctxtoken_rs"] = token

    @staticmethod
    async def afterExecution(
        interface: DispatcherInterface[AvillaEvent],
        exception: Exception | None,
        tb: TracebackType | None,
    ):
        ctx_relationship.reset(interface.local_storage["_ctxtoken_rs"])

    @staticmethod
    async def catch(interface: DispatcherInterface[AvillaEvent]):
        if isinstance(interface.event, AvillaEvent):
            if interface.annotation is Relationship:
                return interface.local_storage["relationship"]


"""
class MetadataDispatcher(BaseDispatcher):
    @staticmethod
    async def catch(interface: DispatcherInterface[AvillaEvent]):
        if isinstance(interface.event, AvillaEvent):
            if isinstance(interface.annotation, type) and issubclass(interface.annotation, Cell):
                relationship: Relationship = interface.local_storage["relationship"]
                return await relationship.meta(interface.annotation)
"""



from __future__ import annotations

from avilla.core.trait import Fn, OrientedFn, Trait
from avilla.core.utilles.selector import Selector


class SceneTrait(Trait):
    @OrientedFn
    async def leave(self) -> None:
        ...

    @OrientedFn
    async def disband(self) -> None:
        ...

    @Fn
    async def remove_member(self, target: Selector, reason: str | None = None) -> None:
        ...

    # TODO: join, invite etc.



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, cast

from graia.amnesia.message import MessageChain

from avilla.core.cell import Cell
from avilla.core.platform import Land
from avilla.core.utilles.selector import Selector

if TYPE_CHECKING:
    from datetime import datetime


@dataclass
class Message(Cell):
    id: str
    mainline: Selector
    sender: Selector
    content: MessageChain
    time: datetime
    reply: Selector | None = None

    @property
    def land(self):
        return Land(cast(str, self.mainline.pattern.get("land")))

    def to_selector(self) -> Selector:
        return self.mainline.copy().message(self.id)



from __future__ import annotations

from datetime import timedelta

from avilla.core.trait import Fn, Trait
from avilla.core.utilles.selector import Selector


class PrivilegeTrait(Trait):
    @Fn
    async def upgrade(self, target: Selector, dest: str | None = None) -> None:
        ...

    @Fn
    async def downgrade(self, target: Selector, dest: str | None = None) -> None:
        ...


class MuteTrait(Trait):
    @Fn
    async def mute(self, target: Selector, duration: timedelta) -> None:
        ...

    @Fn
    async def unmute(self, target: Selector) -> None:
        ...

    @Fn
    async def mute_all(self, target: Selector) -> None:
        ...
    
    @Fn
    async def unmute_all(self, target: Selector) -> None:
        ...

    # Fetch => rs.pull(MuteInfo, target=...)


class BanTrait(Trait):
    @Fn
    async def ban(self, target: Selector, duration: timedelta | None = None) -> None:
        ...

    @Fn
    async def unban(self, target: Selector) -> None:
        ...



from __future__ import annotations

import importlib.metadata
import re
from inspect import cleandoc
from typing import TYPE_CHECKING, Any, cast

from graia.broadcast import Broadcast
from launart import Launart, Service
from loguru import logger

from avilla.core.account import AbstractAccount
from avilla.core.context import get_current_avilla
from avilla.core.dispatchers import AvillaBuiltinDispatcher, RelationshipDispatcher
from avilla.core.platform import Land
from avilla.core.protocol import BaseProtocol
from avilla.core.resource import LocalFileResource
from avilla.core.service import AvillaService
from avilla.core.trait.signature import ResourceFetch
from avilla.core.utilles.selector import Selector

if TYPE_CHECKING:
    from avilla.core.trait.signature import ArtifactSignature

AVILLA_ASCII_LOGO = cleandoc(
    r"""
    [bold]Avilla[/]: a universal asynchronous message flow solution, powered by [blue]Graia Project[/].
        _        _ _ _
       / \__   _(_) | | __ _
      / _ \ \ / / | | |/ _` |
     / ___ \ V /| | | | (_| |
    /_/   \_\_/ |_|_|_|\__,_|
    """
)
AVILLA_ASCII_RAW_LOGO = re.sub(r"\[.*?\]", "", AVILLA_ASCII_LOGO)


GRAIA_PROJECT_REPOS = [
    "avilla-core",
    "graia-broadcast",
    "graia-saya",
    "graia-scheduler",
    "graia-ariadne",
    "statv",
    "launart",
    "creart",
    "creart-graia",
    "kayaku",
]


def _log_telemetry():
    for telemetry in GRAIA_PROJECT_REPOS:
        try:
            version = importlib.metadata.version(telemetry)
        except Exception:
            version = "unknown / not-installed"
        logger.info(
            f"{telemetry} version: {version}",
            alt=f"[b cornflower_blue]{telemetry}[/] version: [cyan3]{version}[/]",
        )


class Avilla:
    broadcast: Broadcast
    launch_manager: Launart
    protocols: list[BaseProtocol]
    # action_middlewares: list[ActionMiddleware]
    # resource_providers: dict[str, ResourceProvider]
    # extension_impls: dict[type[ActionExtension], ActionExtensionImpl]
    accounts: list[AbstractAccount]
    service: AvillaService
    global_artifacts: dict[ArtifactSignature, Any]

    # NOTE: configuration is done by kayaku.
    def __init__(
        self,
        broadcast: Broadcast,
        protocols: list[BaseProtocol],
        services: list[Service],
        # middlewares: list[ActionMiddleware] | None = None,
        launch_manager: Launart | None = None,
    ):
        if len({type(i) for i in protocols}) != len(protocols):
            raise ValueError("protocol must be unique, and the config should be passed by config.")

        self.broadcast = broadcast
        self.launch_manager = launch_manager or Launart()
        self.protocols = protocols
        self._protocol_map = {type(i): i for i in protocols}
        # self.action_middlewares = middlewares or []
        self.accounts = []
        # self.resource_providers = {}
        # self.extension_impls = {}
        self.global_artifacts = {}
        self.service = AvillaService(self)

        for service in services:
            self.launch_manager.add_service(service)

        self.launch_manager.add_service(self.service)

        for protocol in self.protocols:
            # Ensureable 用于注册各种东西, 包括 Service, ResourceProvider 等.
            protocol.ensure(self)

        # self.broadcast.finale_dispatchers.append(MetadataDispatcher())
        self.broadcast.finale_dispatchers.append(AvillaBuiltinDispatcher(self))
        self.broadcast.finale_dispatchers.append(RelationshipDispatcher())

        @self.register_global_artifact(ResourceFetch(LocalFileResource))
        async def _fetch_local_file(_, res: LocalFileResource):
            return res.file.read_bytes()

    @classmethod
    def current(cls) -> "Avilla":
        return get_current_avilla()

    @property
    def loop(self):
        return self.broadcast.loop

    def register_global_artifact(self, signature: ArtifactSignature):
        def warpper(v):
            self.global_artifacts[signature] = v
            return v

        return warpper

    """
    def add_action_middleware(self, middleware: ActionMiddleware):
        self.action_middlewares.append(middleware)

    def remove_action_middleware(self, middleware: ActionMiddleware):
        self.action_middlewares.remove(middleware)

    def get_resource_provider(self, resource: Selector) -> ResourceProvider | None:
        return self.resource_providers.get(cast(str, resource.latest_key))

    def add_resource_provider(self, provider: ResourceProvider, *resource_types: str):
        for resource_type in resource_types:
            self.resource_providers[resource_type] = provider

    def remove_resource_provider(self, provider: ResourceProvider):
        for resource_type in self.resource_providers:
            if self.resource_providers[resource_type] is provider:
                del self.resource_providers[resource_type]

    """

    def add_account(self, account: AbstractAccount):
        if account in self.accounts:
            raise ValueError("account already exists.")
        self.accounts.append(account)

    def remove_account(self, account: AbstractAccount):
        if account not in self.accounts:
            raise ValueError("account not exists.")
        self.accounts.remove(account)

    def get_account(
        self, account_id: str | None = None, selector: Selector | None = None, land: Land | None = None
    ) -> AbstractAccount | None:
        for account in self.accounts:
            if account_id is not None and account.id != account_id:
                continue
            if selector is not None and not selector.match(account.to_selector()):
                continue
            if land is not None and account.land != land:
                continue
            return account

    def get_accounts(
        self, account_id: str | None = None, selector: Selector | None = None, land: Land | None = None
    ) -> list[AbstractAccount]:
        result = []
        for account in self.accounts:
            if account_id is not None and account.id != account_id:
                continue
            if selector is not None and not selector.match(account.to_selector()):
                continue
            if land is not None and account.land != land:
                continue
            result.append(account)
        return result

    async def launch(self):
        logger.info(AVILLA_ASCII_RAW_LOGO, alt=AVILLA_ASCII_LOGO)
        _log_telemetry()

        for protocol in self.protocols:
            if protocol.__class__.platform is not BaseProtocol.platform:
                logger.info(
                    f"Using platform: {protocol.__class__.platform}",
                    alt=f"[magenta]Using platform: [/][dark_orange]{protocol.__class__.platform}[/]",
                )

        await self.launch_manager.launch()



from __future__ import annotations

from pathlib import Path
from typing import Generic, TypeAlias, TypeVar

from avilla.core.cell import Cell
from avilla.core.utilles.selector import Selector

T = TypeVar("T")


class Resource(Cell, Generic[T]):
    selector: Selector

    def __init__(self, selector: Selector):
        self.selector = selector

    def to_selector(self):
        return self.selector


BlobResource: TypeAlias = Resource[bytes]


class LocalFileResource(BlobResource):
    file: Path

    def __init__(self, file: Path | str):
        if isinstance(file, str):
            file = Path(file)
        self.file = file

    @property
    def selector(self) -> Selector:
        return Selector().land("avilla-core").local_file(str(self.file))


from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Generic,
    Protocol,
    TypeVar,
    runtime_checkable,
)

if TYPE_CHECKING:
    from avilla.core.protocol import BaseProtocol

"""
TProtocol = TypeVar("TProtocol", bound="BaseProtocol")


ActionExtensionImpl = Callable[
    ["RelationshipExecutor", "ActionExtension", dict[str, Any] | None], Coroutine[None, None, Any]
]
"""

_T = TypeVar("_T", contravariant=True)


@runtime_checkable
class Ensureable(Protocol, Generic[_T]):
    def ensure(self, interact: _T) -> Any:
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

from avilla.core.request import Request

from . import AvillaEvent

if TYPE_CHECKING:
    from datetime import datetime


class RequestEvent(AvillaEvent):
    request: Request

    @property
    def mainline(self):
        return self.request.mainline

    @property
    def ctx(self):
        return self.request.sender

    def __init__(
        self,
        request: Request,
        time: datetime | None = None,
    ):
        self.request = request
        super().__init__(request.account, time=time)

    class Dispatcher(BaseDispatcher):
        @staticmethod
        async def catch(interface: DispatcherInterface[RequestEvent]):
            if interface.annotation is Request:
                return interface.event.request


class RequestReceived(RequestEvent):
    pass


class RequestAccepted(RequestEvent):
    pass


class RequestRejected(RequestEvent):
    pass


class RequestIgnored(RequestEvent):
    pass


class RequestCancelled(RequestEvent):
    pass



from __future__ import annotations

from collections.abc import Callable, Container, Iterable
from functools import partial
from inspect import isawaitable
from operator import contains, eq, is_, is_not, ne
from typing import Any, Awaitable, Generic, ParamSpec, TypeVar

from graia.broadcast import BaseDispatcher, DispatcherInterface, ExecutionStop
from typing_extensions import Self

from avilla.core.account import AbstractAccount
from avilla.core.relationship import Relationship
from avilla.core.utilles.selector import Selectable, Selector

T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")


class Filter(BaseDispatcher, Generic[T]):
    middlewares: list[Callable[[Any], Awaitable[Any] | Any]]

    def fetch(self, annotation: type[R]) -> Filter[R]:
        @self.middlewares.append
        async def _(_) -> R:
            interface = DispatcherInterface.ctx.get()
            result = await interface.lookup_param("__filter_fetch__", annotation, None)
            interface.local_storage.setdefault("__filter_fetch__", {})[annotation] = result
            return result

        return self  # type: ignore

    def step(self, func: Callable[[T], R]) -> Filter[R]:
        self.middlewares.append(func)
        return self  # type: ignore

    def assert_true(self, func: Callable[[T], Any]) -> Self:
        @self.step
        def _(result: T) -> T:
            if func(result):
                return result
            raise ExecutionStop

        return self

    def assert_false(self, func: Callable[[T], Any]) -> Self:
        return self.assert_true(lambda result: not func(result))

    def assert_equal(self, value: R) -> R:
        return self.assert_true(partial(eq, value))  # type: ignore

    def assert_not_equal(self, value: Any) -> Self:
        return self.assert_true(partial(ne, value))

    def assert_is(self, value: R) -> Filter[R]:
        return self.assert_true(partial(is_, value))  # type: ignore

    def assert_is_not(self, value: Any) -> Self:
        return self.assert_true(partial(is_not, value))

    def assert_is_none(self) -> Filter[None]:
        return self.assert_is(None)

    def assert_is_not_none(self) -> Self:  # 实际上类型是 Filter[T - None]
        return self.assert_is_not(None)

    def assert_in(self, container: Container[R]) -> Filter[R]:
        return self.assert_true(partial(contains, container))  # type: ignore

    def assert_not_in(self, container: Container) -> Self:
        return self.assert_true(lambda result: result not in container)

    def assert_is_instance(self, cls: type[R]) -> Filter[R]:
        return self.assert_true(lambda result: isinstance(result, cls))  # type: ignore

    def assert_not_is_instance(self, cls: type) -> Self:
        return self.assert_true(lambda result: not isinstance(result, cls))

    def any(self, funcs: Iterable[Callable[[T], Any]]) -> Self:
        funcs = frozenset(funcs)
        return self.assert_true(lambda result: any(func(result) for func in funcs))

    def all(self, funcs: Iterable[Callable[[T], Any]]) -> Self:
        funcs = frozenset(funcs)
        return self.assert_true(lambda result: all(func(result) for func in funcs))

    def match(self: Filter[Selectable], pattern: Selector) -> Filter[Selectable]:
        return self.assert_true(lambda result: pattern.match(result))

    def mismatch(self: Filter[Selectable], pattern: Selector) -> Filter[Selectable]:
        return self.assert_false(lambda result: pattern.match(result))

    def match_any(self: Filter[Selectable], patterns: Iterable[Selector]) -> Filter[Selectable]:
        return self.any(lambda result: pattern.match(result) for pattern in patterns)

    def match_all(self: Filter[Selectable], patterns: Iterable[Selector]) -> Filter[Selectable]:
        return self.all(lambda result: pattern.match(result) for pattern in patterns)

    def mismatch_any(self: Filter[Selectable], patterns: Iterable[Selector]) -> Filter[Selectable]:
        return self.any(lambda result: not pattern.match(result) for pattern in patterns)

    def mismatch_all(self: Filter[Selectable], patterns: Iterable[Selector]) -> Filter[Selectable]:
        return self.any(lambda result: not pattern.match(result) for pattern in patterns)

    def follows(self: Filter[Selectable], *patterns: str) -> Filter[Selectable]:
        return self.any(lambda result: result.to_selector().follows(pattern) for pattern in patterns)

    @classmethod
    @property
    def rs(cls) -> Filter[Relationship]:
        return cls().fetch(Relationship)

    @property
    def ctx(self: Filter[Relationship]) -> Filter[Selectable]:
        return self.rs.step(lambda rs: rs.ctx)

    @property
    def mainline(self: Filter[Relationship]) -> Filter[Selector]:
        return self.rs.step(lambda rs: rs.mainline)

    @property
    def current(self: Filter[Relationship]) -> Filter[AbstractAccount]:
        return self.rs.step(lambda rs: rs.account)

    @property
    def via(self: Filter[Relationship]) -> Filter[Selector | None]:
        return self.rs.step(lambda rs: rs.via)

    async def beforeExecution(self, interface: DispatcherInterface):
        result: Awaitable[Any] | Any = interface  # type: ignore
        for middleware in self.middlewares:
            if isawaitable(result := middleware(result)):
                result = await result

    async def catch(self, interface: DispatcherInterface):
        return interface.local_storage.get("__filter_fetch__", {}).get(interface.annotation)



from __future__ import annotations
from typing import Callable, Literal, TypeVar, Union, Protocol, runtime_checkable, Annotated, Generic

MatchRule = Literal['exact', 'exist']
Pattern = Union[str, int, Callable[[Union[str, int]], bool]]


class Selector:
    match_rule: MatchRule = 'exact'

    pattern: dict[str, Pattern]

    def __init__(self, *, match_rule: MatchRule = 'exact'):
        self.match_rule = match_rule
        self.pattern = {}
    
    def __getattr__(self, name: str):
        def wrapper(content: Pattern):
            self.pattern[name] = content
            return self
        return wrapper

    def contains(self, key: str) -> bool:
        return key in self.pattern
    
    __contains__ = contains

    def __getitem__(self, key: str) -> Pattern:
        return self.pattern[key]
    
    @property
    def constant(self) -> bool:
        return all(not callable(v) for v in self.pattern.values())

    @property
    def path(self) -> str:
        return ".".join(self.pattern.keys())

    def match(self, another: Selector) -> bool:
        if self.match_rule == 'exact':
            if self.constant:
                return another.constant and self.pattern == another.pattern
            
        elif self.match_rule == 'exist':
            return set(self.pattern.keys()).issubset(another.pattern.keys())
        else:
            raise ValueError(f'Unknown match rule: {self.match_rule}')


from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from graia.broadcast.entities.dispatcher import BaseDispatcher
from graia.broadcast.entities.event import Dispatchable
from graia.broadcast.interfaces.dispatcher import DispatcherInterface

if TYPE_CHECKING:
    from ..account import AbstractAccount
    from ..application import Avilla


class AvillaLifecycleEvent(Dispatchable):
    """指示有关应用 (Avilla) 的事件."""

    avilla: Avilla

    def __init__(self, avilla: Avilla):
        self.avilla = avilla

    class Dispatcher(BaseDispatcher):
        @classmethod
        async def catch(cls, interface: "DispatcherInterface[AvillaLifecycleEvent]"):
            from ..application import Avilla

            if interface.annotation is Avilla:
                return interface.event.avilla


class ApplicationClosing(AvillaLifecycleEvent):
    """指示 Avilla 正在关闭."""


class ApplicationClosed(AvillaLifecycleEvent):
    """指示 Avilla 已经关闭."""


class ApplicationPreparing(AvillaLifecycleEvent):
    """指示 Avilla 正在准备."""


class ApplicationReady(AvillaLifecycleEvent):
    """指示 Avilla 已准备完毕."""


@dataclass
class AccountStatusChanged(AvillaLifecycleEvent):
    """指示当前账号 (Account) 状态发生改变的事件"""

    account: "AbstractAccount"

    class Dispatcher(BaseDispatcher):
        @classmethod
        async def catch(cls, interface: "DispatcherInterface[AccountStatusChanged]"):
            from ..account import AbstractAccount

            if issubclass(interface.annotation, AbstractAccount):
                return interface.event.account


class AccountAvailable(AccountStatusChanged):
    """指示当前账号处于可用状态."""


class AccountUnavailable(AccountStatusChanged):
    """指示当前账号处于不可用状态."""



from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta

from . import Cell


@dataclass
class Count(Cell):
    current: int
    max: int


@dataclass
class Summary(Cell):
    name: str
    description: str | None


@dataclass
class MuteInfo(Cell):
    muted: bool
    duration: timedelta


@dataclass
class BanInfo(Cell):
    banned: bool
    duration: timedelta


@dataclass
class Privilege(Cell):
    available: bool
    effective: bool


@dataclass
class Nick(Cell):
    name: str
    nickname: str
    badge: str | None


@dataclass
class Reason(Cell):
    content: str


@dataclass
class Comment(Cell):
    content: str


@dataclass
class QuestionItem:
    id: str
    content: str


@dataclass
class Questions(Cell):
    items: list[QuestionItem]


@dataclass
class AnswerItem:
    id: str
    content: str


@dataclass
class Answers(Cell):
    items: list[AnswerItem]



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING

from avilla.core.event import AvillaEvent
from avilla.core.resource import BlobResource
from avilla.core.utilles.selector import Selector

if TYPE_CHECKING:
    from avilla.core.account import AbstractAccount


class ResourceAvailable(AvillaEvent):
    resource: BlobResource

    @property
    def ctx(self):
        return self.resource.selector

    def __init__(
        self,
        resource: BlobResource,
        account: AbstractAccount,
        time: datetime | None = None,
    ):
        self.resource = resource
        super().__init__(account, time=time)


class ResourceUnavailable(AvillaEvent):
    resource: BlobResource

    @property
    def ctx(self):
        return self.resource.selector

    def __init__(
        self,
        resource: BlobResource,
        account: AbstractAccount,
        time: datetime | None = None,
    ):
        self.resource = resource
        super().__init__(account, time=time)


class FileUploaded(ResourceAvailable):
    uploader: Selector
    mainline: Selector

    def __init__(
        self,
        resource: BlobResource,
        uploader: Selector,
        mainline: Selector,
        account: AbstractAccount,
        time: datetime | None = None,
    ):
        self.uploader = uploader
        self.mainline = mainline
        super().__init__(resource, account, time)


class FileRemoved(ResourceUnavailable):
    uploader: Selector
    mainline: Selector

    def __init__(
        self,
        resource: BlobResource,
        uploader: Selector,
        mainline: Selector,
        account: AbstractAccount,
        time: datetime | None = None,
    ):
        self.uploader = uploader
        self.mainline = mainline
        super().__init__(resource, account, time)



from __future__ import annotations

import abc
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Concatenate,
    Generic,
    ParamSpec,
    Protocol,
    TypeVar,
    overload,
)

from typing_extensions import Unpack

from avilla.core.resource import Resource
from avilla.core.trait import DirectFn, OrientedFn
from avilla.core.trait.signature import CompleteRule
from avilla.core.trait.signature import Impl as _Impl
from avilla.core.trait.signature import ImplDefaultTarget
from avilla.core.trait.signature import Pull as _Pull
from avilla.core.trait.signature import Query
from avilla.core.trait.signature import ResourceFetch as _ResourceFatch
from avilla.core.utilles.selector import Selector

from .context import ctx_prefix, eval_dotpath, get_current_namespace

if TYPE_CHECKING:
    from avilla.core.cell import Cell, CellOf
    from avilla.core.relationship import Relationship
    from avilla.core.trait.signature import ArtifactSignature

    from . import Fn


class Recorder(abc.ABC):
    @abc.abstractmethod
    def signature(self) -> ArtifactSignature:
        ...

    def __call__(self, content: Any):
        sig = self.signature()
        r = get_current_namespace()
        r[sig] = content
        return content


_P = ParamSpec("_P")
_T = TypeVar("_T")


class ImplRecorder(Recorder, Generic[_P, _T]):
    trait_call: Fn
    path: type[Cell] | CellOf | None = None
    target: str | None = None

    @overload
    def __new__(cls, trait_call: OrientedFn[_P, _T]) -> ImplRecorder[Concatenate[Selector, _P], _T]:
        ...

    @overload
    def __new__(cls, trait_call: DirectFn[_P, _T]) -> ImplRecorder[Concatenate[Selector, _P], _T]:
        ...

    @overload
    def __new__(cls, trait_call: Fn[_P, _T]) -> ImplRecorder[_P, _T]:
        ...

    def __new__(cls, trait_call: Fn[_P, _T]) -> ImplRecorder[_P, _T] | ImplRecorder[Concatenate[Selector, _P], _T]:
        return super(ImplRecorder, cls).__new__(cls)

    def __init__(self, trait_call: Fn[_P, _T]):
        self.trait_call = trait_call

    def of(self, path: type[Cell] | CellOf):
        self.path = path
        return self

    def pin(self, target: str):
        self.target = eval_dotpath(target, ctx_prefix.get())
        return self

    def signature(self):
        if isinstance(self.trait_call, DirectFn):  # type: ignore
            target = eval_dotpath(".", ctx_prefix.get())
        else:
            target = self.target
        return _Impl(target=target, path=self.path, trait_call=self.trait_call)  # type: ignore

    def __call__(self, content: Callable[Concatenate[Relationship, _P], Awaitable[_T]]):
        return super().__call__(content)


impl = ImplRecorder

_R = TypeVar("_R", bound=Resource)


class FetchRecorder(Recorder, Generic[_R]):
    resource: type[_R]

    def __init__(self, resource: type[_R]):
        self.resource = resource

    def signature(self):
        return _ResourceFatch(self.resource)

    def __call__(self, content: Callable[[Relationship, _R], Awaitable[Any]]):
        return super().__call__(content)


fetch = FetchRecorder

_C = TypeVar("_C", bound="Cell")


class PullRecorder(Recorder, Generic[_C]):
    target: str | None = None
    path: type[Cell] | CellOf

    def __init__(self, path: type[_C] | CellOf[Unpack[tuple[Any, ...]], _C]):
        self.path = path

    def of(self, target: str):
        self.target = target
        return self

    def signature(self):
        return _Pull(eval_dotpath(self.target, ctx_prefix.get()) if self.target is not None else None, self.path)

    def __call__(self, content: Callable[[Relationship, Selector | None], Awaitable[_C]]):
        return super().__call__(content)


pull = PullRecorder


def completes(relative: str, output: str):
    r = get_current_namespace()
    r[CompleteRule(eval_dotpath(relative, ctx_prefix.get()))] = eval_dotpath(output, ctx_prefix.get())



Querier = Callable[["Relationship", Selector | None, Selector], AsyncGenerator[Selector, None]]


@overload
def query(upper: None, target: str) -> Callable[[Callable[["Relationship", None, Selector], AsyncGenerator[Selector, None]]], Callable[["Relationship", None, Selector], AsyncGenerator[Selector, None]]]:
    ...

@overload
def query(upper: str, target: str) -> Callable[[Callable[["Relationship", Selector, Selector], AsyncGenerator[Selector, None]]], Callable[["Relationship", Selector, Selector], AsyncGenerator[Selector, None]]]:
    ...

def query(upper: str | None, target: str) -> ...:
    def wrapper(querier):
        r = get_current_namespace()
        r[Query(upper, target)] = querier
        return querier
    return wrapper


class ImplDefaultTargetRecorder(Recorder):
    # trait_call: TraitCall
    path: type[Cell] | CellOf | None = None

    def __init__(self, trait_call: Fn):
        self.trait_call = trait_call

    def of(self, path: type[Cell] | CellOf):
        self.path = path
        return self

    def signature(self):
        return ImplDefaultTarget(self.path, self.trait_call)

    def __call__(self, content: Callable[[Relationship], Selector]):
        return super().__call__(content)


default_target = ImplDefaultTargetRecorder
"""
def event_key(func: Callable[[Any], str]):
    r = get_current_namespace()
    r[EventKeyGetter()] = func
    return func

def event(key: str):
    def wrapper(func: Callable[[Any], Avilla])
"""



from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Any

from graia.broadcast.utilles import Ctx

if TYPE_CHECKING:
    from graia.broadcast.entities.event import Dispatchable

    from avilla.core.application import Avilla
    from avilla.core.protocol import BaseProtocol
    from avilla.core.relationship import Relationship


ctx_avilla: Ctx[Avilla] = Ctx("avilla")
ctx_protocol: Ctx[BaseProtocol] = Ctx("protocol")
ctx_relationship: Ctx[Relationship] = Ctx("relationship")


def get_current_avilla() -> Avilla:
    avilla = ctx_avilla.get()
    if avilla:
        return avilla
    protocol = ctx_protocol.get()
    if protocol:
        return protocol.avilla
    relationship = ctx_relationship.get()
    if relationship:
        return relationship.protocol.avilla
    raise RuntimeError("no any current avilla")


def get_current_protocol():
    protocol = ctx_protocol.get()
    if protocol:
        return protocol
    relationship = ctx_relationship.get()
    if relationship:
        return relationship.protocol
    raise RuntimeError("no any current protocol")


def get_current_relationship():
    relationship = ctx_relationship.get()
    if relationship:
        return relationship
    raise RuntimeError("no any current relationship")


def require_relationship(func):
    @functools.wraps(func)
    async def wrapper(*args, **kwargs):
        if ctx_relationship.get():
            return await func(*args, **kwargs)
        raise RuntimeError("no any current relationship")

    return wrapper



from __future__ import annotations

from avilla.core.trait import OrientedFn, Trait


class SummaryTrait(Trait):
    @OrientedFn
    async def set_name(self, name: str) -> None:
        ...

    @OrientedFn
    async def unset_name(self) -> None:
        ...

    @OrientedFn
    async def set_description(self, description: str) -> None:
        ...

    @OrientedFn
    async def unset_description(self) -> None:
        ...



from graia.amnesia.message import MessageChain as MessageChain

from .account import AbstractAccount as AbstractAccount
from .application import Avilla as Avilla
from .cell import Cell as Cell
from .cell.cells import AnswerItem as AnswerItem
from .cell.cells import Answers as Answers
from .cell.cells import BanInfo as BanInfo
from .cell.cells import Comment as Comment
from .cell.cells import Count as Count
from .cell.cells import MuteInfo as MuteInfo
from .cell.cells import Nick as Nick
from .cell.cells import Privilege as Privilege
from .cell.cells import QuestionItem as QuestionItem
from .cell.cells import Questions as Questions
from .cell.cells import Reason as Reason
from .cell.cells import Summary as Summary
from .context import ctx_avilla as ctx_avilla
from .context import ctx_protocol as ctx_protocol
from .context import ctx_relationship as ctx_relationship
from .context import get_current_avilla as get_current_avilla
from .context import get_current_protocol as get_current_protocol
from .context import get_current_relationship as get_current_relationship
from .context import require_relationship as require_relationship
from .dispatchers import AvillaBuiltinDispatcher as AvillaBuiltinDispatcher
from .dispatchers import RelationshipDispatcher as RelationshipDispatcher
from .elements import Audio as Audio
from .elements import Notice as Notice
from .elements import NoticeAll as NoticeAll
from .elements import Picture as Picture
from .elements import Text as Text
from .elements import Unknown as Unknown
from .elements import Video as Video
from .event import AvillaEvent as AvillaEvent
from .event import MetadataModified as MetadataModified
from .event import RelationshipCreated as RelationshipCreated
from .event import RelationshipDestroyed as RelationshipDestroyed
from .event.lifecycle import AccountAvailable as AccountAvailable
from .event.lifecycle import AccountStatusChanged as AccountStatusChanged
from .event.lifecycle import AccountUnavailable as AccountUnavailable
from .event.lifecycle import ApplicationClosed as ApplicationClosed
from .event.lifecycle import ApplicationClosing as ApplicationClosing
from .event.lifecycle import ApplicationPreparing as ApplicationPreparing
from .event.lifecycle import ApplicationReady as ApplicationReady
from .event.lifecycle import AvillaLifecycleEvent as AvillaLifecycleEvent
from .event.message import MessageEdited as MessageEdited
from .event.message import MessageReceived as MessageReceived
from .event.message import MessageRevoked as MessageRevoked
from .event.request import RequestAccepted as RequestAccepted
from .event.request import RequestCancelled as RequestCancelled
from .event.request import RequestEvent as RequestEvent
from .event.request import RequestIgnored as RequestIgnored
from .event.request import RequestReceived as RequestReceived
from .event.request import RequestRejected as RequestRejected
from .event.resource import ResourceAvailable as ResourceAvailable
from .event.resource import ResourceUnavailable as ResourceUnavailable
from .exceptions import AccountDeleted as AccountDeleted
from .exceptions import AccountMuted as AccountMuted
from .exceptions import ActionFailed as ActionFailed
from .exceptions import ContextError as ContextError
from .exceptions import DeprecatedError as DeprecatedError
from .exceptions import HttpRequestError as HttpRequestError
from .exceptions import InaccessibleInterface as InaccessibleInterface
from .exceptions import InvalidAuthentication as InvalidAuthentication
from .exceptions import InvalidOperation as InvalidOperation
from .exceptions import NetworkError as NetworkError
from .exceptions import ParserException as ParserException
from .exceptions import RemoteError as RemoteError
from .exceptions import TooLongMessage as TooLongMessage
from .exceptions import UnknownError as UnknownError
from .exceptions import UnknownTarget as UnknownTarget
from .exceptions import UnsupportedOperation as UnsupportedOperation
from .message import Message as Message
from .platform import Abstract as Abstract
from .platform import Branch as Branch
from .platform import Land as Land
from .platform import Maintainer as Maintainer
from .platform import Platform as Platform
from .platform import PlatformDescription as PlatformDescription
from .platform import Version as Version
from .protocol import BaseProtocol as BaseProtocol
from .querier import AbstractQueryHandler as AbstractQueryHandler
from .querier import ProtocolAbstractQueryHandler as ProtocolAbstractQueryHandler
from .relationship import Relationship as Relationship
from .request import Request as Request
from .resource import LocalFileResource as LocalFileResource
from .resource import Resource as Resource
from .service import AvillaService as AvillaService
from .typing import Ensureable as Ensureable
from .utilles.selector import DynamicSelector as DynamicSelector
from .utilles.selector import Selectable as Selectable
from .utilles.selector import Selector as Selector



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Hashable

if TYPE_CHECKING:
    from avilla.core.cell import Cell, CellOf
    from avilla.core.resource import Resource
    from avilla.core.trait import Fn


class ArtifactSignature:
    pass


@dataclass(unsafe_hash=True)
class Impl(ArtifactSignature):
    trait_call: Fn
    target: str | None
    path: type[Cell] | CellOf | None


@dataclass(unsafe_hash=True)
class Pull(ArtifactSignature):
    target: str | None
    path: type[Cell] | CellOf


@dataclass(unsafe_hash=True)
class ResourceFetch(ArtifactSignature):
    target: type[Resource]


@dataclass(unsafe_hash=True)
class Allow(ArtifactSignature):
    content: Hashable


@dataclass(unsafe_hash=True)
class CompleteRule(ArtifactSignature):
    relative: str


@dataclass(unsafe_hash=True)
class ImplDefaultTarget(ArtifactSignature):
    path: type[Cell] | CellOf | None
    trait_call: Fn


@dataclass(unsafe_hash=True)
class Query(ArtifactSignature):
    upper: str | None
    target: str

@dataclass(unsafe_hash=True)
class Check(ArtifactSignature):
    target: str


from __future__ import annotations

from avilla.core.event import AvillaEvent
from avilla.core.utilles.selector import Selector


class ActivityAvailable(AvillaEvent):
    id: Selector  # eg. [...].activity("button_clicked")
    activity: Selector # eg. [...].button("#1")

    @property
    def ctx(self):
        return self.activity

class ActivityUnavailable(AvillaEvent):
    id: Selector
    activity: Selector

    @property
    def ctx(self):
        return self.activity

class ActivityTrigged(AvillaEvent):
    id: Selector
    activity: Selector

    trigger: Selector  # who trigged the activity
    mainline: Selector
    
    @property
    def ctx(self):
        return self.trigger